@@ -4,7 +4,7 @@
 #history http://www.reportlab.co.uk/cgi-bin/viewcvs.cgi/public/reportlab/trunk/reportlab/lib/_rl_accel.c
  ****************************************************************************/
 #if 0
-static __version__=" $Id: _rl_accel.c 2580 2005-04-25 14:08:26Z rgbecker $ "
+static __version__=" $Id: _rl_accel.c 2924 2006-06-16 13:33:10Z rgbecker $ "
 #endif
 #include "Python.h"
 #include <stdlib.h>
@@ -27,11 +27,12 @@ static __version__=" $Id: _rl_accel.c 2580 2005-04-25 14:08:26Z rgbecker $ "
 #ifndef min
 #	define min(a,b) ((a)<(b)?(a):(b))
 #endif
-#define VERSION "0.53"
+#define VERSION "0.57"
 #define MODULE "_rl_accel"
 
-
 static PyObject *moduleVersion;
+static PyObject *moduleObject;
+static int moduleLineno;
 typedef struct _fI_t {
 		char*			name;
 		int				ascent, descent;
@@ -619,7 +620,7 @@ static PyObject *_instanceEscapePDF(PyObject *unused, PyObject* args)
 static PyObject *_sameFrag(PyObject *self, PyObject* args)
 {
 	PyObject *f, *g;
-	static char *names[] = {"fontName", "fontSize", "textColor", "rise", "underline", NULL};
+	static char *names[] = {"fontName", "fontSize", "textColor", "rise", "underline", "strike", "link", NULL};
 	int	r=0, t;
 	char **p;
 	if (!PyArg_ParseTuple(args, "OO:_sameFrag", &f, &g)) return NULL;
@@ -627,22 +628,24 @@ static PyObject *_sameFrag(PyObject *self, PyObject* args)
 	for(p=names;*p;p++){
 		PyObject *fa, *ga;
 		fa = PyObject_GetAttrString(f,*p);
-		if(!fa){
-L1:			return NULL;
-			}
 		ga = PyObject_GetAttrString(g,*p);
-		if(!ga){
+		if(fa && ga){
+			t = PyObject_Compare(fa,ga);
 			Py_DECREF(fa);
-			goto L1;
+			Py_DECREF(ga);
+			if(PyErr_Occurred()) goto L1;
+			}
+		else{
+			t = fa==ga ? 0 : 1;
+			Py_XDECREF(fa);
+			Py_XDECREF(ga);
+			PyErr_Clear();
 			}
-		t = PyObject_Compare(fa,ga);
-		Py_DECREF(fa);
-		Py_DECREF(ga);
-		if(PyErr_Occurred()) goto L1;
 		if(t) goto L0;
 		}
 	r = 1;
 L0:	return PyInt_FromLong((long)r);
+L1:	return NULL;
 }
 
 static PyObject *ttfonts_calcChecksum(PyObject *self, PyObject* args)
@@ -727,6 +730,556 @@ static PyObject *hex32(PyObject *self, PyObject* args)
 	return PyString_FromString(buf);
 }
 
+#if PY_VERSION_HEX>=0x02030000
+static PyObject *_notdefFont=NULL;
+static PyObject *_notdefChar=NULL;
+
+static PyObject *_GetExcValue(void)
+{
+	PyObject *type = NULL, *value = NULL, *tb = NULL, *result=NULL;
+	PyErr_Fetch(&type, &value, &tb);
+	PyErr_NormalizeException(&type, &value, &tb);
+	if(PyErr_Occurred()) goto L_BAD;
+	if(!value){
+		value = Py_None;
+		Py_INCREF(value);
+		}
+	Py_XINCREF(value);
+	result = value;
+L_BAD:
+	Py_XDECREF(type);
+	Py_XDECREF(value);
+	Py_XDECREF(tb);
+	return result;
+}
+static PyObject *_GetAttrString(PyObject *obj, char *name)
+{
+	PyObject *res = PyObject_GetAttrString(obj, name);
+	if(!res) PyErr_SetString(PyExc_AttributeError, name);
+	return res;
+}
+
+#if 0
+#	define ERROR_EXIT() goto L_ERR;
+#	define ADD_TB(name)
+#else
+#	define ERROR_EXIT() {moduleLineno=__LINE__;goto L_ERR;}
+#	define ADD_TB(name) _add_TB(name)
+#endif
+#include "compile.h"
+#include "frameobject.h"
+#include "traceback.h"
+static void _add_TB(char *funcname)
+{
+	PyObject *py_srcfile = NULL, *py_funcname = NULL, *py_globals = NULL, *empty_tuple = NULL, *empty_string = NULL;
+	PyCodeObject *py_code = NULL;
+	PyFrameObject *py_frame = NULL;
+	
+	py_srcfile = PyString_FromString(__FILE__);
+	if(!py_srcfile) goto bad;
+	py_funcname = PyString_FromString(funcname);
+	if(!py_funcname) goto bad;
+	py_globals = PyModule_GetDict(moduleObject);
+	if(!py_globals) goto bad;
+	empty_tuple = PyTuple_New(0);
+	if(!empty_tuple) goto bad;
+	empty_string = PyString_FromString("");
+	if(!empty_string) goto bad;
+	py_code = PyCode_New(
+						0,				/*int argcount,*/
+						0,				/*int nlocals,*/
+						0,				/*int stacksize,*/
+						0,				/*int flags,*/
+						empty_string,	/*PyObject *code,*/
+						empty_tuple,	/*PyObject *consts,*/
+						empty_tuple,	/*PyObject *names,*/
+						empty_tuple,	/*PyObject *varnames,*/
+						empty_tuple,	/*PyObject *freevars,*/
+						empty_tuple,	/*PyObject *cellvars,*/
+						py_srcfile,		/*PyObject *filename,*/
+						py_funcname,	/*PyObject *name,*/
+						moduleLineno,	/*int firstlineno,*/
+						empty_string	/*PyObject *lnotab*/
+						);
+	if(!py_code) goto bad;
+	py_frame = PyFrame_New(
+		PyThreadState_Get(), /*PyThreadState *tstate,*/
+		py_code,			 /*PyCodeObject *code,*/
+		py_globals,			 /*PyObject *globals,*/
+		0					 /*PyObject *locals*/
+		);
+	if(!py_frame) goto bad;
+	py_frame->f_lineno = moduleLineno;
+	PyTraceBack_Here(py_frame);
+bad:
+	Py_XDECREF(py_srcfile);
+	Py_XDECREF(py_funcname);
+	Py_XDECREF(empty_tuple);
+	Py_XDECREF(empty_string);
+	Py_XDECREF(py_code);
+	Py_XDECREF(py_frame);
+}
+static PyObject *unicode2T1(PyObject *self, PyObject *args, PyObject *kwds)
+{
+	int			i, j, _i1, _i2;
+	PyObject	*R, *font, *enc, *res, *utext=NULL, *fonts=NULL,
+				*_o1 = NULL, *_o2 = NULL, *_o3 = NULL;
+	static char *argnames[] = {"utext","fonts",NULL};
+	char		*encStr;
+	if (!PyArg_ParseTupleAndKeywords(args, kwds, "OO", argnames, &utext, &fonts)) return NULL;
+	Py_INCREF(utext);
+	Py_INCREF(fonts);
+	R = Py_None; Py_INCREF(Py_None);
+	font = Py_None; Py_INCREF(Py_None);
+	enc = Py_None; Py_INCREF(Py_None);
+
+	if(!_notdefFont){
+		_o1 = PyImport_ImportModule("reportlab.pdfbase.pdfmetrics"); if(!_o1) ERROR_EXIT();
+		_o2 = _GetAttrString(_o1,"_notdefFont");
+		_o3 = _GetAttrString(_o1,"_notdefChar");
+		if(!_o2 || !_o3 || !_o3) ERROR_EXIT();
+		_notdefFont = _o2;
+		_notdefChar = _o3;
+		Py_DECREF(_o1);
+		_o1 = _o2 = _o3 = NULL;
+		}
+
+	_o2 = PyList_New(0); if(!_o2) ERROR_EXIT();
+	Py_DECREF(R);
+	R = _o2;
+	_o2 = NULL;
+
+	_o2 = PySequence_GetItem(fonts,0); if(!_o2) ERROR_EXIT();
+	_o1 = PySequence_GetSlice(fonts, 1, 0x7fffffff); if(!_o1) ERROR_EXIT();
+	Py_DECREF(font);
+	font = _o2;
+	Py_DECREF(fonts);
+	fonts = _o1;
+	_o1 = _o2 = NULL;
+
+	_o2 = _GetAttrString(font, "encName"); if(!_o2) ERROR_EXIT();
+	Py_DECREF(enc);
+	enc = _o2;
+	_o2 = NULL;
+
+	encStr = PyString_AsString(enc);
+	if(PyErr_Occurred()) ERROR_EXIT();
+	if(strstr(encStr,"UCS-2")) encStr = "UTF16";
+
+	while((_i1=PyObject_IsTrue(utext))>0){
+		if((_o1 = PyUnicode_AsEncodedString(utext, encStr, NULL))){
+			_o2 = PyTuple_New(2); if(!_o2) ERROR_EXIT();
+			Py_INCREF(font);
+			PyTuple_SET_ITEM(_o2, 0, font);
+			PyTuple_SET_ITEM(_o2, 1, _o1);
+			_o1 = NULL;
+			if(PyList_Append(R, _o2)) ERROR_EXIT();
+			Py_DECREF(_o2); _o2 = NULL;
+			break;
+			}
+		else{
+			Py_XDECREF(_o1); _o1 = NULL;
+
+			if(!PyErr_ExceptionMatches(PyExc_UnicodeEncodeError)) ERROR_EXIT();
+			_o1 = _GetExcValue(); if(!_o1) ERROR_EXIT();
+			PyErr_Clear();
+			_o2 = _GetAttrString(_o1, "args"); if(!_o2) ERROR_EXIT();
+			Py_DECREF(_o1);
+			_o1 = PySequence_GetSlice(_o2, 2, 4); if(!_o1) ERROR_EXIT();
+			Py_DECREF(_o2);
+				_o2 = PySequence_GetItem(_o1, 0); if(!_o2) ERROR_EXIT();
+				i = PyInt_AsLong(_o2); if(PyErr_Occurred()) ERROR_EXIT();
+				Py_DECREF(_o2);
+
+				_o2 = PySequence_GetItem(_o1, 1); if(!_o1) ERROR_EXIT();
+				j = PyInt_AsLong(_o2); if(PyErr_Occurred()) ERROR_EXIT();
+				Py_DECREF(_o2);
+
+			Py_DECREF(_o1); _o2 = _o1 = 0;
+
+			if(i){
+				_o1 = PySequence_GetSlice(utext, 0, i); if(!_o1) ERROR_EXIT();
+				_o2 = PyUnicode_AsEncodedString(_o1, encStr, NULL); if(!_o2) ERROR_EXIT();
+				Py_DECREF(_o1);
+				_o1 = PyTuple_New(2); if(!_o1) ERROR_EXIT();
+				Py_INCREF(font);
+				PyTuple_SET_ITEM(_o1, 0, font);
+				PyTuple_SET_ITEM(_o1, 1, _o2);
+				_o2 = NULL;
+				if(PyList_Append(R, _o1)) ERROR_EXIT();
+				Py_DECREF(_o1); _o1 = NULL;
+				}
+
+			_i2 = PyObject_IsTrue(fonts); if(_i2<0) ERROR_EXIT();
+			if(_i2){
+				_o1 = PySequence_GetSlice(utext, i, j); if(!_o1) ERROR_EXIT();
+				_o2 = PyTuple_New(2); if(!_o2) ERROR_EXIT();
+				PyTuple_SET_ITEM(_o2, 0, _o1);
+				Py_INCREF(fonts);
+				PyTuple_SET_ITEM(_o2, 1, fonts);
+				_o1 = unicode2T1(self,_o2,NULL); if(!_o1) ERROR_EXIT();
+				Py_DECREF(_o2); _o2 = 0;
+				_o3 = PyTuple_New(1); if(!_o3) ERROR_EXIT();
+				PyTuple_SET_ITEM(_o3, 0, _o1);
+				_o1 = _GetAttrString(R, "extend"); if(!_o1) ERROR_EXIT();
+				_o2 = PyObject_CallObject(_o1, _o3); if(!_o2) ERROR_EXIT();
+				Py_DECREF(_o1);
+				Py_DECREF(_o3);
+				Py_DECREF(_o2); _o1 = _o2 = _o3 = NULL;
+				}
+			else{
+				_o2 = PyInt_FromLong((j - i)); if(!_o2) ERROR_EXIT();
+				_o1 = PyNumber_Multiply(_notdefChar, _o2); if(!_o1) ERROR_EXIT();
+				Py_DECREF(_o2);
+				_o2 = PyTuple_New(2); if(!_o2) ERROR_EXIT();
+				PyTuple_SET_ITEM(_o2, 0, _notdefFont);
+				PyTuple_SET_ITEM(_o2, 1, _o1);
+				Py_INCREF(_notdefFont);
+				_o1 = NULL;
+				if(PyList_Append(R, _o2)) ERROR_EXIT();
+				Py_DECREF(_o2); _o2 = NULL;
+				}
+
+			_o1 = PySequence_GetSlice(utext, j, 0x7fffffff); if(!_o1) ERROR_EXIT();
+			Py_DECREF(utext);
+			utext = _o1;
+			_o1 = NULL;
+			}
+		}
+	if(_i1<0) ERROR_EXIT();
+
+	Py_INCREF(R);
+	res = R;
+	goto L_OK;
+
+L_ERR:
+	ADD_TB("unicode2T1");
+	Py_XDECREF(_o1);
+	Py_XDECREF(_o2);
+	Py_XDECREF(_o3);
+	res = NULL;
+L_OK:
+	Py_DECREF(R);
+	Py_DECREF(font);
+	Py_DECREF(enc);
+	Py_DECREF(utext);
+	Py_DECREF(fonts);
+	return res;
+}
+static PyObject *_pdfmetrics_fonts = NULL;	/*the fontName to font map from pdfmetrics*/
+static PyObject *_pdfmetrics_ffar = NULL;	/*findFontAndRegister from pdfmetrics*/
+static PyObject *getFontU(PyObject *module, PyObject *args, PyObject *kwds)
+{
+	PyObject *fontName=NULL, *_o1=NULL, *_o2=NULL, *res=NULL;
+	static char *argnames[] = {"fontName",NULL};
+	if(!PyArg_ParseTupleAndKeywords(args, kwds, "O", argnames, &fontName)) return NULL;
+	if(!_pdfmetrics_fonts){
+		res = PyImport_ImportModule("reportlab.pdfbase.pdfmetrics");
+		if(!res) ERROR_EXIT();
+		_o1 = _GetAttrString(res,"_fonts");
+		if(!_o1) ERROR_EXIT();
+		_o2 = _GetAttrString(res,"findFontAndRegister");
+		if(!_o2) ERROR_EXIT();
+		_pdfmetrics_fonts = _o1;
+		_pdfmetrics_ffar = _o2;
+		Py_DECREF(res); _o1 = _o2 = res = NULL;
+		}
+	if((res = PyObject_GetItem(_pdfmetrics_fonts,fontName))) return res;
+	if(!PyErr_ExceptionMatches(PyExc_KeyError)) ERROR_EXIT();
+	PyErr_Clear();
+	_o1 = PyTuple_New(1); if(!_o1) ERROR_EXIT();
+	PyTuple_SET_ITEM(_o1, 0, fontName);
+	Py_INCREF(fontName);
+	res = PyObject_CallObject(_pdfmetrics_ffar,_o1);
+	Py_DECREF(_o1);	/*NB this should decremnent fontName as well*/
+	return res;
+
+L_ERR:
+	ADD_TB("getFontU");
+	Py_XDECREF(_o1);
+	Py_XDECREF(_o2);
+	Py_XDECREF(res);
+	return NULL;
+}
+static PyObject *stringWidthU(PyObject *self, PyObject *args, PyObject *kwds)
+{
+	PyObject	*text=NULL, *fontName=NULL, *fontSize=NULL, *encoding=NULL,
+				*res=NULL, *_o1=NULL, *_o2=NULL, *_o3=NULL;
+	static char *argnames[] = {"text","fontName","fontSize","encoding",0};
+	if(!PyArg_ParseTupleAndKeywords(args, kwds, "OOO|O", argnames, &text, &fontName, &fontSize, &_o1)) return NULL;
+	Py_INCREF(text);
+	Py_INCREF(fontName);
+	Py_INCREF(fontSize);
+	if(_o1){
+		encoding = _o1;
+		_o1 = NULL;
+		Py_INCREF(encoding);
+		}
+	else{
+		_o1 = PyString_FromString("utf8"); if(!_o1) ERROR_EXIT();
+		encoding = _o1;
+		_o1 = NULL;
+		}
+
+	/*getFont(fontName).stringWidth --> _o2*/
+	_o1 = PyTuple_New(1); if(!_o1) ERROR_EXIT();
+	Py_INCREF(fontName);
+	PyTuple_SET_ITEM(_o1, 0, fontName);
+	_o2 = getFontU(self,_o1,NULL); if(!_o2) ERROR_EXIT();
+	Py_DECREF(_o1);
+	_o1 = _GetAttrString(_o2, "stringWidth"); if(!_o1) ERROR_EXIT();
+	Py_DECREF(_o2);
+
+	_o2 = PyTuple_New(2); if(!_o2) ERROR_EXIT();
+	Py_INCREF(text);
+	PyTuple_SET_ITEM(_o2, 0, text);
+	Py_INCREF(fontSize);
+	PyTuple_SET_ITEM(_o2, 1, fontSize);
+	_o3 = PyDict_New(); if(!_o3) ERROR_EXIT();
+	if(PyDict_SetItemString(_o3, "encoding", encoding)<0) ERROR_EXIT();
+	res = PyEval_CallObjectWithKeywords(_o1, _o2, _o3); if(!res) ERROR_EXIT();
+	Py_DECREF(_o1);
+	Py_DECREF(_o2);
+	Py_DECREF(_o3);
+	goto L_OK;
+L_ERR:
+	ADD_TB("stringWidthU");
+	Py_XDECREF(_o1);
+	Py_XDECREF(_o2);
+	Py_XDECREF(_o3);
+L_OK:
+	Py_DECREF(text);
+	Py_DECREF(fontName);
+	Py_DECREF(fontSize);
+	Py_DECREF(encoding);
+	return res;
+}
+static PyObject *_instanceStringWidthU(PyObject *module, PyObject *args, PyObject *kwds)
+{
+	PyObject *L, *t, *f, *self, *text, *size, *res,
+				*encoding = 0, *_o1 = 0, *_o2 = 0, *_o3 = 0;
+	unsigned char *b;
+	int n, m, i, j, s, _i1;
+	static char *argnames[]={"self","text","size","encoding",0};
+	if(!PyArg_ParseTupleAndKeywords(args, kwds, "OOO|O", argnames, &self, &text, &size, &_o1)) return 0;
+	Py_INCREF(self);
+	Py_INCREF(text);
+	Py_INCREF(size);
+	if(_o1){
+		encoding = _o1;
+		_o1 = NULL;
+		Py_INCREF(encoding);
+		}
+	else{
+		_o1 = PyString_FromString("utf8"); if(!_o1) ERROR_EXIT();
+		encoding = _o1;
+		_o1 = NULL;
+		}
+	L = Py_None; Py_INCREF(Py_None);
+	t = Py_None; Py_INCREF(Py_None);
+	f = Py_None; Py_INCREF(Py_None);
+
+	if(!PyUnicode_Check(text)){
+		_o1 = _GetAttrString(text, "decode"); if(!_o1) ERROR_EXIT();
+		_o3 = PyTuple_New(1); if(!_o3) ERROR_EXIT();
+		Py_INCREF(encoding);
+		PyTuple_SET_ITEM(_o3, 0, encoding);
+		_o2 = PyObject_CallObject(_o1, _o3); if(!_o2) ERROR_EXIT();
+		Py_DECREF(_o1);
+		Py_DECREF(_o3); _o1 = _o3 = NULL;
+		Py_DECREF(text);
+		text = _o2;
+		_o2 = NULL;
+		}
+
+	_o3 = PyList_New(1); if(!_o3) ERROR_EXIT();
+	Py_INCREF(self);
+	PyList_SET_ITEM(_o3, 0, self);
+	_o2 = _GetAttrString(self, "substitutionFonts"); if(!_o2) ERROR_EXIT();
+	_o1 = PyNumber_Add(_o3, _o2); if(!_o1) ERROR_EXIT();
+	Py_DECREF(_o3); _o3 = 0;
+	Py_DECREF(_o2); _o2 = NULL;
+	_o3 = PyTuple_New(2); if(!_o3) ERROR_EXIT();
+	Py_INCREF(text);
+	PyTuple_SET_ITEM(_o3, 0, text);
+	PyTuple_SET_ITEM(_o3, 1, _o1);
+	_o1 = NULL;
+	_o2 = unicode2T1(module,_o3,NULL); if(!_o2) ERROR_EXIT();
+	Py_DECREF(_o3); _o3 = NULL;
+	Py_DECREF(L);
+	L = _o2;
+	_o2 = NULL;
+
+	n = PyList_GET_SIZE(L);
+
+	for(s=i=0;i<n;++i){
+		_o1 = PyList_GetItem(L,i); if(!_o1) ERROR_EXIT();
+		Py_INCREF(_o1);
+
+		_o2 = PySequence_GetItem(_o1, 0); if(!_o2) ERROR_EXIT();
+		Py_DECREF(f);
+		f = _o2;
+		_o2 = NULL;
+
+		_o2 = _GetAttrString(f, "widths"); if(!_o2) ERROR_EXIT();
+		Py_DECREF(f);
+		f = _o2;
+		_o2 = NULL;
+
+		_o2 = PySequence_GetItem(_o1, 1); if(!_o2) ERROR_EXIT();
+		Py_DECREF(t);
+		t = _o2;
+		Py_DECREF(_o1);
+		_o1 = _o2 = NULL;
+
+		m = PyString_Size(t);
+		b = PyString_AS_STRING(t);
+
+		for(j=0;j<m;++j){
+			_i1 = (long)(b[j]);
+			_o2 = PyList_GetItem(f,_i1); if(!_o2) {PyErr_Format(PyExc_IndexError,"widths index %d out of range",_i1);ERROR_EXIT();}
+			_i1 = PyInt_AsLong(_o2);
+			_o2 = NULL;	/*we borrowed this*/
+			if(PyErr_Occurred()) ERROR_EXIT();
+			s += _i1;
+			}
+		}
+
+	_o1 = PyFloat_FromDouble((s * 0.001)); if(!_o1) ERROR_EXIT();
+	res = PyNumber_Multiply(_o1, size); if(!res) ERROR_EXIT();
+	Py_DECREF(_o1);
+	goto L_OK;
+L_ERR:
+	ADD_TB("_instanceStringWidthU");
+	Py_XDECREF(_o1);
+	Py_XDECREF(_o2);
+	Py_XDECREF(_o3);
+	res = NULL;
+L_OK:
+	Py_DECREF(L);
+	Py_DECREF(t);
+	Py_DECREF(f);
+	Py_DECREF(self);
+	Py_DECREF(text);
+	Py_DECREF(size);
+	Py_DECREF(encoding);
+	return res;
+}
+static PyObject *_instanceStringWidthTTF(PyObject *module, PyObject *args, PyObject *kwds)
+{
+	PyObject *self, *text, *size, *res,
+				*encoding = 0, *_o1=NULL, *_o2=NULL, *_o3=NULL;
+	Py_UNICODE *b;
+	int n, i;
+	double s, _d1, dw;
+	static char *argnames[]={"self","text","size","encoding",0};
+	if(!PyArg_ParseTupleAndKeywords(args, kwds, "OOO|O", argnames, &self, &text, &size, &_o1)) return 0;
+	Py_INCREF(text);
+	if(_o1){
+		encoding = _o1;
+		_o1 = NULL;
+		Py_INCREF(encoding);
+		}
+	else{
+		_o1 = PyString_FromString("utf8"); if(!_o1) ERROR_EXIT();
+		encoding = _o1;
+		_o1 = NULL;
+		}
+
+	if(!PyUnicode_Check(text)){
+		i = PyObject_IsTrue(encoding); if(i<0) ERROR_EXIT();
+		if(!i){
+			Py_DECREF(encoding);
+			encoding = PyString_FromString("utf8"); if(!encoding) ERROR_EXIT();
+			}
+		_o1 = _GetAttrString(text, "decode"); if(!_o1) ERROR_EXIT();
+		_o3 = PyTuple_New(1); if(!_o3) ERROR_EXIT();
+		Py_INCREF(encoding);
+		PyTuple_SET_ITEM(_o3, 0, encoding);
+		_o2 = PyObject_CallObject(_o1, _o3); if(!_o2) ERROR_EXIT();
+		Py_DECREF(_o1);
+		Py_DECREF(_o3); _o1 = _o3 = NULL;
+		Py_DECREF(text);
+		text = _o2; /*no _o2=NULL as we assign there straight away*/ 
+		}
+
+	/*self.face.charWidths --> _o1, self.face.defaultWidth --> _o3*/
+	_o2 = _GetAttrString(self, "face"); if(!_o2) ERROR_EXIT();
+	_o1 = _GetAttrString(_o2, "charWidths"); if(!_o1) ERROR_EXIT(); if(!PyDict_Check(_o1)){PyErr_SetString(PyExc_TypeError, "TTFontFace instance charWidths is not a dict");ERROR_EXIT();}
+	_o3 = _GetAttrString(_o2, "defaultWidth"); if(!_o3) ERROR_EXIT();
+	Py_DECREF(_o2); _o2 = NULL;
+	dw = PyFloat_AsDouble(_o3);
+	if(PyErr_Occurred()) ERROR_EXIT();
+	Py_DECREF(_o3);	_o3=NULL;
+
+	n = PyUnicode_GET_SIZE(text);
+	b = PyUnicode_AS_UNICODE(text);
+
+	for(s=i=0;i<n;++i){
+		_o3 = PyInt_FromLong((long)b[i]); if(!_o3) ERROR_EXIT();
+		_o2 = PyDict_GetItem(_o1,_o3);
+		Py_DECREF(_o3); _o3 = NULL;
+		if(!_o2) _d1 = dw;
+		else{
+			_d1 = PyFloat_AsDouble(_o2);
+			_o2=NULL;	/*no decref as we borrowed it*/
+			if(PyErr_Occurred()) ERROR_EXIT();
+			}
+		s += _d1;
+		}
+	Py_DECREF(_o1);
+	_o1 = PyFloat_FromDouble((s * 0.001)); if(!_o1) ERROR_EXIT();
+	res = PyNumber_Multiply(_o1, size); if(!res) ERROR_EXIT();
+	Py_DECREF(_o1);
+	goto L_OK;
+L_ERR:
+	ADD_TB("_instanceStringWidthTTF");
+	Py_XDECREF(_o1);
+	Py_XDECREF(_o2);
+	Py_XDECREF(_o3);
+	res = NULL;
+L_OK:
+	Py_DECREF(text);
+	Py_DECREF(encoding);
+	return res;
+}
+/*we may need to reload pdfmtrics etc etc*/
+static PyObject *_reset(PyObject *module)
+{
+	if(_notdefFont){
+		Py_DECREF(_notdefFont); _notdefFont = NULL;
+		Py_DECREF(_notdefChar);	_notdefChar = NULL;
+		}
+	if(_pdfmetrics_fonts){
+		Py_DECREF(_pdfmetrics_fonts); _pdfmetrics_fonts = NULL;
+		Py_DECREF(_pdfmetrics_ffar); _pdfmetrics_ffar = NULL;
+		}
+	Py_INCREF(Py_None);
+	return Py_None;
+}
+#if 0
+/*cannot seem to speed this up*/
+static PyObject *_instanceGetCharWidth(PyObject *module, PyObject *args, PyObject *kwds)
+{
+	PyObject *self, *code, *res, *_o1=NULL;
+	static char *argnames[]={"self","code",0};
+	if(!PyArg_ParseTupleAndKeywords(args, kwds, "OO", argnames, &self, &code)) return 0;
+	_o1 = _GetAttrString(self, "charWidths"); if(!_o1) ERROR_EXIT(); if(!PyDict_Check(_o1)){PyErr_SetString(PyExc_TypeError, "TTFontFace instance charWidths is not a dict");ERROR_EXIT();}
+	res = PyDict_GetItem(_o1,code);
+	if(!res){
+		res = _GetAttrString(self, "defaultWidth"); if(!res) ERROR_EXIT();
+		}
+	Py_INCREF(res);
+	Py_DECREF(_o1);
+	return res;
+L_ERR:
+	ADD_TB("_instanceGetCharWidth");
+	Py_XDECREF(_o1);
+	return NULL;
+}
+#endif
+#endif
+
 #if PY_VERSION_HEX>=0x02030000
 #define HAVE_BOX
 /*Box start**************/
@@ -1033,13 +1586,13 @@ static PyTypeObject BoxList_type = {
 
 static char *__doc__=
 "_rl_accel contains various accelerated utilities\n\
-\tstringWidth a fast string width function\n\
-\t_instanceStringWidth a method version of stringWidth\n\
-\tdefaultEncoding gets/sets the default encoding for stringWidth\n\
-\tgetFonts gets font names from the internal table\n\
-\tgetFontInfo gets font info from the internal table\n\
-\tsetFontInfo adds a font to the internal table\n\
-\t_SWRecover gets/sets a callback for stringWidth recovery\n\
+\tstringWidth a fast string width function (legacy)\n\
+\t_instanceStringWidth a method version of stringWidth (legacy)\n\
+\tdefaultEncoding gets/sets the default encoding for stringWidth (legacy)\n\
+\tgetFonts gets font names from the internal table (legacy)\n\
+\tgetFontInfo gets font info from the internal table (legacy)\n\
+\tsetFontInfo adds a font to the internal table (legacy)\n\
+\t_SWRecover gets/sets a callback for stringWidth recovery (legacy)\n\
 \tescapePDF makes a string safe for PDF\n\
 \t_instanceEscapePDF method equivalent of escapePDF\n\
 \n\
@@ -1050,6 +1603,14 @@ static char *__doc__=
 "\tcalcChecksum calculate checksums for TTFs\n"
 "\tadd32 32 bit unsigned addition\n"
 "\thex32 32 bit unsigned to 0X8.8X string\n"
+#if PY_VERSION_HEX>=0x02030000
+"\tstringWidthU version2 stringWidth\n\
+\t_instanceStringWidthU version2 Font instance stringWidth\n\
+\t_instanceStringWidthTTF version2 TTFont instance stringWidth\n\
+\tgetFontU version2 pdfdmetrics.getFont\n\
+\tunicodeT1 version2 pdfmetrics.unicode2T1\n\
+\t_reset() version2 clears _rl_accel state\n"
+#endif
 #ifdef	HAVE_BOX
 "\tBox(width,character=None) creates a Knuth character Box with the specified width.\n"
 "\tGlue(width,stretch,shrink) creates a Knuth glue Box with the specified width, stretch and shrink.\n"
@@ -1079,6 +1640,17 @@ static struct PyMethodDef _methods[] = {
 	{"calcChecksum", ttfonts_calcChecksum, METH_VARARGS, "calcChecksum(string) calculate checksums for TTFs"},
 	{"add32", ttfonts_add32, METH_VARARGS, "add32(x,y)  32 bit unsigned x+y"},
 	{"hex32", hex32, METH_VARARGS, "hex32(x)  32 bit unsigned-->0X8.8X string"},
+#if PY_VERSION_HEX>=0x02030000
+	{"unicode2T1", (PyCFunction)unicode2T1, METH_VARARGS|METH_KEYWORDS, "return a list of (font,string) pairs representing the unicode text"},
+	{"getFontU", (PyCFunction)getFontU, METH_VARARGS|METH_KEYWORDS, "getFontU(name)-->Font instance"},
+	{"stringWidthU", (PyCFunction)stringWidthU, METH_VARARGS|METH_KEYWORDS, "stringWidthU(text,fontName,fontSize,encoding='utf8')--> font stringWidth(text,fontSize,encoding)"},
+	{"_instanceStringWidthU", (PyCFunction)_instanceStringWidthU, METH_VARARGS|METH_KEYWORDS, "Font.stringWidth(self,text,fontName,fontSize,encoding='utf8') --> width"},
+	{"_instanceStringWidthTTF", (PyCFunction)_instanceStringWidthTTF, METH_VARARGS|METH_KEYWORDS, "TTFont.stringWidth(self,text,fontName,fontSize,encoding='utf8') --> width"},
+	{"_reset", (PyCFunction)_reset, METH_NOARGS, "_rl_accel._reset() reset _rl_accel state"},
+#if	0
+	{"_instanceGetCharWidth", (PyCFunction)_instanceGetCharWidth, METH_VARARGS|METH_KEYWORDS, "TTFontFace.getCharWidth(self,code) --> width"},
+#endif
+#endif
 #ifdef	HAVE_BOX
 	{"Box",	(PyCFunction)Box,	METH_VARARGS|METH_KEYWORDS, "Box(width,character=None) create a Knuth Box instance"},
 	{"Glue", (PyCFunction)Glue,	METH_VARARGS|METH_KEYWORDS, "Glue(width,stretch,shrink) create a Knuth Glue instance"},
@@ -1090,13 +1662,12 @@ static struct PyMethodDef _methods[] = {
 /*Initialization function for the module (*must* be called init_pdfmetrics)*/
 void init_rl_accel(void)
 {
-	PyObject *m;
 #if PY_VERSION_HEX<0x02000000
 	PyObject *d;
 #endif
 
 	/*Create the module and add the functions and module doc string*/
-	m = Py_InitModule3("_rl_accel", _methods,__doc__);
+	moduleObject = Py_InitModule3("_rl_accel", _methods,__doc__);
 
 	/*Add some symbolic constants to the module */
 	if(!ErrorObject){
@@ -1106,10 +1677,10 @@ void init_rl_accel(void)
 	Py_INCREF(ErrorObject);
 	moduleVersion = PyString_FromString(VERSION);
 #if PY_VERSION_HEX>=0x02000000
-	PyModule_AddObject(m, "error", ErrorObject);
-	PyModule_AddObject(m, "version", moduleVersion );
+	PyModule_AddObject(moduleObject, "error", ErrorObject);
+	PyModule_AddObject(moduleObject, "version", moduleVersion );
 #else
-	d = PyModule_GetDict(m);
+	d = PyModule_GetDict(moduleObject);
 	PyDict_SetItemString(d, "error", ErrorObject );
 	PyDict_SetItemString(d, "version", moduleVersion );
 #endif
@@ -1119,7 +1690,7 @@ void init_rl_accel(void)
 	BoxList_type.tp_base = &PyList_Type;
 	if(PyType_Ready(&BoxList_type)<0) goto err;
 	Py_INCREF(&BoxList_type);
-	if(PyModule_AddObject(m, "BoxList", (PyObject *)&BoxList_type)<0)goto err;
+	if(PyModule_AddObject(moduleObject, "BoxList", (PyObject *)&BoxList_type)<0)goto err;
 #endif
 
 err:/*Check for errors*/
